@@ -45,8 +45,7 @@ pub use crate::extern_ref::{VMExternObj, VMExternRef};
 pub use crate::function_env::VMFunctionEnvironment;
 pub use crate::global::*;
 pub use crate::imports::Imports;
-#[allow(deprecated)]
-pub use crate::instance::{InstanceAllocator, InstanceHandle, VMInstance};
+pub use crate::instance::{InstanceAllocator, VMInstance};
 pub use crate::memory::{
     initialize_memory_with_data, LinearMemory, NotifyLocation, VMMemory, VMOwnedMemory,
     VMSharedMemory,
@@ -72,12 +71,6 @@ use wasmer_types::RawValue;
 pub use wasmer_types::TableStyle;
 pub use wasmer_types::{StoreId, TargetSharedSignatureIndex, VMBuiltinFunctionIndex, VMOffsets};
 
-#[deprecated(
-    since = "2.1.0",
-    note = "ModuleInfo, ExportsIterator, ImportsIterator should be imported from wasmer_types."
-)]
-pub use wasmer_types::{ExportsIterator, ImportsIterator, ModuleInfo};
-
 /// Version number of this crate.
 pub const VERSION: &str = env!("CARGO_PKG_VERSION");
 
